#ifndef STRUCT_DVB
#define STRUCT_DVB
extern char *dvb_fe_status[32];
extern char *dvb_fe_caps[32];

extern char *dvb_fe_type[];
extern char *dvb_fe_bandwidth[];
extern char *dvb_fe_rates[];
extern char *dvb_fe_modulation[];
extern char *dvb_fe_transmission[];
extern char *dvb_fe_guard[];
extern char *dvb_fe_hierarchy[];
extern char *dvb_fe_inversion[];

#endif
